#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (c) 2021 Kazuhiro KOBAYASHI <root.4mac@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""

import joblib
import torch
import torch.nn as nn

from .layer import (Conv1d, ConvLayers, ConvLTVFilterGenerator,
                    SinusoidsGenerator)
from .preprocess_layer import (LogMelSpectrogram2LogMagnitude,
                               LogMelSpectrogramScaler)


class NeuralHomomorphicVocoder(nn.Module):
    def __init__(
        self,
        fs=24000,
        fft_size=1024,
        hop_size=256,
        in_channels=80,
        conv_channels=256,
        ltv_out_channels=222,
        out_channels=1,
        kernel_size=3,
        group_size=8,
        dilation_size=1,
        fmin=80,
        fmax=7600,
        roll_size=24,
        look_ahead=32,
        use_causal=False,
        use_ddsconv=False,
        use_tanh=False,
        use_uvmask=True,
        use_conv_postfilter=False,
        use_ddsconv_pf=True,
        use_ltv_conv_postfilter=False,
        use_reference_mag=False,
        use_quefrency_norm=True,
        use_weight_norm=False,
        use_clip_grad_norm=False,
        scaler_file=None,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.use_conv_postfilter = use_conv_postfilter
        self.use_tanh = use_tanh

        if scaler_file is not None:
            self.melspc_scaler_fn = LogMelSpectrogramScaler(
                joblib.load(scaler_file)["mlfb"]
            )
        else:
            self.melspc_scaler_fn = None

        if use_reference_mag:
            mel2linear_fn = LogMelSpectrogram2LogMagnitude(
                fs=fs,
                fft_size=fft_size,
                n_mels=in_channels,
                fmin=fmin,
                fmax=fmax,
                roll_size=roll_size,
                melspc_scaler_fn=self.melspc_scaler_fn,
            )
        else:
            mel2linear_fn = None

        ltv_params = {
            "in_channels": in_channels,
            "conv_channels": conv_channels,
            "ltv_out_channels": ltv_out_channels,
            "kernel_size": kernel_size,
            "dilation_size": dilation_size,
            "group_size": group_size,
            "fft_size": fft_size,
            "hop_size": hop_size,
            "use_causal": use_causal,
            "use_ddsconv": use_ddsconv,
            "use_ltv_conv_postfilter": use_ltv_conv_postfilter,
            "look_ahead": look_ahead,
            "use_quefrency_norm": use_quefrency_norm,
        }
        self.impuluse_generator = SinusoidsGenerator(
            hop_size=hop_size, fs=fs, use_uvmask=use_uvmask
        )
        self.ltv_harmonic = ConvLTVFilterGenerator(
            **ltv_params, mel2spc_fn=mel2linear_fn
        )
        self.ltv_noise = ConvLTVFilterGenerator(**ltv_params)

        if self.use_conv_postfilter:
            if use_ddsconv_pf:
                self.conv_pf = ConvLayers(
                    in_channels=1,
                    conv_channels=64,
                    out_channels=1,
                    kernel_size=5,
                    dilation_size=2,
                    group_size=8,
                    use_causal=True,
                    look_ahead=hop_size,
                    use_quefrency_norm=False,
                )
            else:
                self.conv_pf = Conv1d(
                    in_channels=1,
                    out_channels=1,
                    kernel_size=fft_size,
                    use_causal=True,
                    look_ahead=hop_size,
                )

        if use_weight_norm:
            self.apply_weight_norm()

        if use_clip_grad_norm:
            self.apply_clip_grad_norm()

    def _forward(self, x, cf0, uv):
        """
        x: (B, T, D)
        cf0: (B, T, 1)
        uv: (B, T, 1)
        """
        if self.melspc_scaler_fn is not None:
            x = self.melspc_scaler_fn(x)

        harmonic, noise = self.impuluse_generator(cf0, uv)
        sig_harm = self.ltv_harmonic(x, harmonic)
        sig_noise = self.ltv_noise(x, noise)
        y = sig_harm + sig_noise

        if self.use_conv_postfilter:
            y = self.conv_pf(y.transpose(1, 2)).transpose(1, 2)

        if self.use_tanh:
            y = torch.tanh(y)

        return y.reshape(x.size(0), self.out_channels, -1)

    def forward(self, z, c):
        """Interface for PWG trainer
        z: (B, T, D)
        c: (B, T, n_mels + 2)
        """
        c = c.transpose(1, 2)
        x, cf0, uv = torch.split(c, [self.in_channels, 1, 1], dim=-1)
        y = self._forward(x, cf0, uv)
        y = torch.clip(y, -1, 1)
        return y

    @torch.no_grad()
    def inference(self, c):
        """Interface for PWG decoder
        c: (T, D)
        """
        c = c.unsqueeze(0)
        x, cf0, uv = torch.split(c, [self.in_channels, 1, 1], dim=-1)
        y = self._forward(x, cf0, uv)
        return y.squeeze(0)

    def remove_weight_norm(self):
        def _remove_weight_norm(m):
            try:
                torch.nn.utils.remove_weight_norm(m)
            except ValueError:
                return

        self.apply(_remove_weight_norm)

    def apply_weight_norm(self):
        def _apply_weight_norm(m):
            if isinstance(m, torch.nn.Conv1d):
                torch.nn.utils.weight_norm(m)

        self.apply(_apply_weight_norm)

    def apply_clip_grad_norm(self, max_norm=3):
        torch.nn.utils.clip_grad_norm_(self.parameters(), max_norm=max_norm)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (c) 2021 Kazuhiro KOBAYASHI <root.4mac@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""

from pathlib import Path

import pytest
import soundfile as sf
import torch
import torch.nn.functional as F

B, T, D = 3, 100, 80
hop_size, fft_size, window_size = 128, 1024, 1024
fs = 24000

dirname = Path(__file__).parent
scalerf = dirname / "data" / "test_scaler.pkl"


@pytest.mark.parametrize(
    [
        "use_causal",
        "use_ddsconv",
        "use_conv_postfilter",
        "use_ltv_conv_postfilter",
        "use_reference_mag",
        "use_tanh",
        "use_uvmask",
        "use_ddsconv_pf",
        "scaler_file",
    ],
    [
        (True, False, False, False, False, False, False, False, None),
        (False, False, False, False, False, False, False, False, None),
        (False, False, True, False, False, False, False, False, None),
        (False, False, False, True, False, False, False, False, None),
        (False, False, False, False, True, False, False, False, None),
        (False, False, False, False, False, False, False, False, scalerf),
        (False, True, True, True, False, False, False, False, scalerf),
        (True, True, True, True, False, False, False, False, scalerf),
        (True, True, True, True, False, True, False, False, scalerf),
        (True, True, True, True, False, True, True, False, scalerf),
        (True, True, True, True, False, True, True, True, scalerf),
    ],
    ids=[
        "use_causal",
        "not_use_causal",
        "use_conv_postfilter",
        "use_ltv_conv_postfilter",
        "use_reference_mag",
        "use_scaler",
        "use_ddsconv",
        "use_causal_ddsconv",
        "use_tanh",
        "use_uvmask",
        "use_ddsconv_pf"
    ],
)
def test_nhv(
    use_causal,
    use_ddsconv,
    use_conv_postfilter,
    use_ltv_conv_postfilter,
    use_reference_mag,
    use_tanh,
    use_uvmask,
    use_ddsconv_pf,
    scaler_file,
):
    from nhv.layer import NeuralHomomorphicVocoder

    net = NeuralHomomorphicVocoder(
        fs=fs,
        fft_size=fft_size,
        hop_size=hop_size,
        in_channels=D,
        use_causal=use_causal,
        use_ddsconv=use_ddsconv,
        use_tanh=use_tanh,
        use_uvmask=use_uvmask,
        use_conv_postfilter=use_conv_postfilter,
        use_ddsconv_pf=use_ddsconv_pf,
        use_ltv_conv_postfilter=use_ltv_conv_postfilter,
        use_reference_mag=use_reference_mag,
        scaler_file=scaler_file,
    )
    z = torch.randn((B, 1, T * hop_size))
    x = torch.randn((B, T, D))
    f0 = torch.randn(B, T, 1)
    uv = torch.randn(B, T, 1)
    y = net._forward(x, f0, uv)  # noqa
    y = net.forward(z, torch.cat([x, f0, uv], dim=-1).transpose(1, 2))
    assert y.size(2) == T * hop_size


def test_layer():
    from nhv.layer import ConvLTVFilterGenerator

    x = torch.randn((B, T, D))
    z = torch.randn((B, 1, T * hop_size))
    conv = ConvLTVFilterGenerator(
        in_channels=80, ltv_out_channels=256, fft_size=fft_size, hop_size=hop_size
    )
    y = conv(x, z)  # noqa


def test_sinusoids_generator():
    from nhv.layer import SinusoidsGenerator

    net = SinusoidsGenerator(hop_size=hop_size, fs=fs)
    cf0 = torch.arange(100, 100 + T).reshape(1, -1, 1)
    cf0 = torch.cat([cf0, cf0, cf0], axis=0).float()
    uv = torch.ones_like(cf0).float()
    excit, noise = net(cf0, uv)
    outf = dirname / "test_sinusoids_generator_output.wav"
    sf.write(outf, excit[0].squeeze().numpy(), fs)
    outf.unlink()


def test_sinusoids_generator_fromf0():
    import numpy as np
    from nhv.layer import SinusoidsGenerator

    net = SinusoidsGenerator(hop_size=hop_size, fs=fs)
    f0 = np.loadtxt(dirname / "data" / "test.f0")
    cf0 = torch.from_numpy(f0).reshape(1, -1, 1)
    cf0 = torch.cat([cf0, cf0], axis=0)
    uv = torch.ones_like(cf0).float()
    excit, noise = net(cf0, uv)
    outf = dirname / "test_sinusoids_generator_from_f0_output.wav"
    sf.write(outf, excit[0].squeeze().numpy(), fs)
    outf.unlink()


def test_unfold():
    x = torch.randn(B, T, hop_size)
    _ = torch.randn(B, T, fft_size)

    z = torch.randn(B, 1, T * hop_size)
    z = z.squeeze()  # B, T x hop_size
    z = F.pad(z, (fft_size, fft_size - 1))
    z = z.unfold(-1, fft_size * 2, step=hop_size)  # B x T x window_size
    z = F.pad(z, (hop_size // 2, hop_size // 2 - 1))
    z = z.unfold(-1, hop_size, step=1)  # B x T x window_size x hop_size
    z = torch.matmul(z.squeeze(), x.unsqueeze(-1)).squeeze()  # B x T x window_size


def test_ola():
    win = torch.bartlett_window(window_size, periodic=False)
    z = torch.randn(3, 100, window_size)
    z = z * win
    l, r = torch.chunk(z, 2, dim=-1)
    z = l + r.roll(1, dims=-1)
    z = z.reshape(z.size(0), -1).unsqueeze(1)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (c) 2021 Kazuhiro KOBAYASHI <root.4mac@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""


import matplotlib as mpl
import numpy.testing as npt
import torch

mpl.use("Agg")
import matplotlib.pyplot as plt  # noqa

B = 3


def test_interpolate():
    x = torch.randn((1, 10, 513))
    mask = x < 0
    x[mask] = -100


def test_ccep_calculation():
    window_size = 1024
    fft_size = window_size
    win = torch.hann_window(window_size)
    x = torch.randn(B, 1, window_size) * win

    # to fft->ifft
    y = torch.fft.fft(x, n=fft_size, dim=-1)
    x_ret = torch.fft.ifft(y, n=fft_size, dim=-1).real
    npt.assert_almost_equal(x.numpy(), x_ret.numpy(), decimal=5)

    # win -> logmag, phase -> ccep -> logmag, phase -> win
    log_mag, phase = torch.log(torch.abs(y)), torch.angle(y)
    comp = torch.complex(log_mag, phase)
    ccep = torch.fft.ifft(comp, n=fft_size, dim=-1)
    logmagphase = torch.fft.fft(ccep, n=fft_size, dim=-1)
    magphase = torch.exp(logmagphase)
    x_ret2 = torch.fft.ifft(magphase, n=fft_size, dim=-1).real
    npt.assert_almost_equal(x.numpy(), x_ret2.numpy(), decimal=5)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (c) 2021 Kazuhiro KOBAYASHI <root.4mac@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""

from pathlib import Path

import matplotlib as mpl
import soundfile as sf
import torch

mpl.use("agg")
import matplotlib.pyplot as plt  # noqa

dirname = Path(__file__).parent
wavf = dirname / "data" / "test.wav"


def test_melspectrogram():
    from nhv.layer import LogMelSpectrogram

    x, fs = sf.read(wavf)

    hop_size = 128
    fft_size = 1024
    win_length = 1024
    window = "hann"
    n_mels = 80
    fmin = None
    fmax = None
    mel_layer = LogMelSpectrogram(
        fs=fs,
        hop_size=hop_size,
        fft_size=fft_size,
        win_length=win_length,
        window=window,
        n_mels=n_mels,
        fmin=fmin,
        fmax=fmax,
        center=True,
        pad_mode="reflect",
    )
    x = torch.from_numpy(x).unsqueeze(0).float()
    x = mel_layer(x)


def test_melspec2logspec():
    from nhv.layer import LogMelSpectrogram2LogMagnitude  # noqa
    from nhv.layer import CepstrumLiftering, LogMelSpectrogram, Magnitude

    x, fs = sf.read(wavf)
    x = torch.from_numpy(x).unsqueeze(0).float()

    hop_size = 128
    fft_size = 1024
    win_length = 1024
    window = "hann"
    n_mels = 80
    fmin = 80
    fmax = 7600

    mag_layer = Magnitude(
        fs=fs,
        hop_size=hop_size,
        fft_size=fft_size,
        win_length=win_length,
        window=window,
    )
    mag = mag_layer(x)
    log_mag = torch.clamp(mag, min=1e-10).log10()

    log_mel_layer = LogMelSpectrogram(
        fs=fs,
        hop_size=hop_size,
        fft_size=fft_size,
        win_length=win_length,
        window=window,
        n_mels=n_mels,
        fmin=fmin,
        fmax=fmax,
        center=True,
        pad_mode="reflect",
    )
    inv_mel_layer = LogMelSpectrogram2LogMagnitude(
        fs=fs,
        fft_size=fft_size,
        n_mels=n_mels,
        fmin=fmin,
        fmax=fmax,
        eps=1.0e-10,
        roll_size=16,
    )
    log_melspc = log_mel_layer(x)
    log_mag_inv = inv_mel_layer(log_melspc)

    cep_layer = CepstrumLiftering(lifter_size=24)
    log_mag_inv_lifter = cep_layer(log_mag_inv)

    T, D = 100, 512
    plt.figure()
    plt.plot(log_mag_inv.squeeze().cpu().numpy()[T, :D])
    plt.plot(log_mag.squeeze().cpu().numpy()[T, :D])
    plt.plot(log_mag_inv_lifter.squeeze().cpu().numpy()[T, :D])
    plt.savefig(dirname / "data" / "test_melspec2logspec.png")

    plt.figure()
    plt.plot(log_melspc.squeeze().cpu().numpy()[T])
    plt.savefig(dirname / "data" / "test_log_melspc.png")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (c) 2021 Kazuhiro KOBAYASHI <root.4mac@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""

import librosa
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


class LogMelSpectrogram(nn.Module):
    def __init__(
        self,
        fs=24000,
        hop_size=128,
        fft_size=1024,
        win_length=None,
        window="hann",
        center=True,
        pad_mode="reflect",
        n_mels=80,
        fmin=None,
        fmax=None,
        scaler_file=None,
    ):
        super().__init__()
        self.mag_layer = Magnitude(fs, hop_size, fft_size, win_length, window)
        self.mel_layer = Magnitude2LogMelSpectrogram(fs, fft_size, n_mels, fmin, fmax)
        if scaler_file is not None:
            self.melspc_scaler = LogMelSpectrogramScaler(scaler_file)
        else:
            self.melspc_scaler = None

    def forward(self, x):
        mag = self.mag_layer(x)
        log_melspc = self.mel_layer(mag)
        if self.melspc_scaler is not None:
            log_melspc = self.scaler_layer(log_melspc)
        return log_melspc


class Magnitude(torch.nn.Module):
    def __init__(
        self,
        fs=24000,
        hop_size=128,
        fft_size=1024,
        win_length=None,
        window="hann",
        center=True,
        pad_mode="reflect",
        return_complex=True,
    ):
        super().__init__()
        self.hop_size = hop_size
        self.fft_size = fft_size
        self.win_length = fft_size if win_length is None else win_length
        self.window = window
        self.center = center
        self.pad_mode = pad_mode
        self.return_complex = return_complex

    def forward(self, x):
        """
        x: (B, 1, T)
        ret: (B, T, fft_size // 2 + 1)
        """
        f = getattr(torch, f"{self.window}_window")
        window = f(self.win_length, dtype=x.dtype, device=x.device)
        y = torch.stft(
            x,
            n_fft=self.fft_size,
            win_length=self.win_length,
            hop_length=self.hop_size,
            window=window,
            center=self.center,
            pad_mode=self.pad_mode,
            return_complex=self.return_complex,
        )
        return y.abs().transpose(1, 2)


class Magnitude2LogMelSpectrogram(torch.nn.Module):
    def __init__(
        self, fs=24000, fft_size=1024, n_mels=80, fmin=None, fmax=None, eps=1.0e-10
    ):
        super().__init__()
        self.eps = eps
        fmin = 0 if fmin is None else fmin
        fmax = fs / 2 if fmax is None else fmax
        mel_basis = librosa.filters.mel(
            fs, fft_size, n_mels=n_mels, fmin=fmin, fmax=fmax
        )
        self.register_buffer("mel_basis", torch.from_numpy(mel_basis.T).float())

    def forward(self, x):
        melspc = torch.matmul(x, self.mel_basis)
        log_melspc = torch.clamp(melspc, min=self.eps).log10()
        return log_melspc


class LogMelSpectrogram2LogMagnitude(nn.Module):
    def __init__(
        self,
        fs,
        fft_size,
        n_mels=80,
        fmin=None,
        fmax=None,
        eps=1.0e-10,
        roll_size=24,
        melspc_scaler_fn=None,
    ):
        super().__init__()
        self.eps = eps
        self.roll_size = roll_size
        self.melspc_scaler_fn = melspc_scaler_fn
        fmin = 0 if fmin is None else fmin
        fmax = fs / 2 if fmax is None else fmax

        mel_basis = librosa.filters.mel(
            fs, fft_size, n_mels=n_mels, fmin=fmin, fmax=fmax
        )
        inv_mel_basis = np.linalg.pinv(mel_basis)
        self.register_buffer("inv_mel_basis", torch.from_numpy(inv_mel_basis.T).float())

    def forward(self, x):
        if self.melspc_scaler_fn is not None:
            # denorm mlfb
            x = self.melspc_scaler_fn.inverse_transform(x)

        x = torch.pow(10.0, x)
        spc = torch.matmul(x, self.inv_mel_basis)
        log_spc = 10 * torch.clamp(spc, min=self.eps).log10()
        z = F.pad(log_spc, (self.roll_size // 2 - 1, self.roll_size // 2))
        z = z.unfold(-1, self.roll_size, step=1)
        log_spc = torch.median(z, dim=-1)[0]
        return log_spc


class CepstrumLiftering(nn.Module):
    def __init__(self, lifter_size=None):
        super().__init__()
        if lifter_size <= 0:
            raise ValueError("lifter_size must be > 0.")
        else:
            self.lifter_size = lifter_size

    def forward(self, x):
        cep = torch.fft.ifft(x, dim=-1)
        cep[..., self.lifter_size : -self.lifter_size] = 0
        x = torch.fft.fft(cep, dim=-1)
        return x


class LogMelSpectrogramScaler(nn.Module):
    def __init__(self, scaler):
        super().__init__()
        self.register_parameter(
            "mean",
            nn.Parameter(torch.from_numpy(scaler.mean_).float(), requires_grad=False),
        )
        self.register_parameter(
            "scale",
            nn.Parameter(
                torch.from_numpy(scaler.var_).float().sqrt(), requires_grad=False
            ),
        )

    def forward(self, x):
        return (x - self.mean) / self.scale

    def inverse_transform(self, x):
        return x * self.scale + self.mean



#!/usr/bin/env python

import os

from setuptools import find_packages, setup

if os.path.exists("README.md"):
    README = open("README.md").read()
else:
    README = ""  # a placeholder, readme is generated on release
CHANGES = open("CHANGES.md").read()

setup(
    name="neural-homomorphic-vocoder",
    version="0.0.7",
    description="Pytorch implementation of neural homomorphic vocoder",
    url="https://github.com/k2kobayashi/neural-homomorphic-vocoder",
    author="K. KOBAYASHI",
    packages=find_packages(exclude=["examples"]),
    long_description=(README + "\n" + CHANGES),
    long_description_content_type='text/markdown',
    license="MIT",
    install_requires=open("tools/requirements.txt").readlines(),
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright (c) 2021 Kazuhiro KOBAYASHI <root.4mac@gmail.com>
#
# Distributed under terms of the MIT license.

"""

"""

import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class Conv1d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        dilation_size=1,
        group_size=1,
        use_causal=False,
        look_ahead=0,
    ):
        super().__init__()
        self.use_causal = use_causal
        self.look_ahead = look_ahead
        self.kernel_size = kernel_size
        self.padding = (kernel_size - 1) * dilation_size

        self.conv1d = nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size,
            padding=self.padding,
            dilation=dilation_size,
            groups=group_size,
        )
        nn.init.kaiming_normal_(self.conv1d.weight)

        if use_causal and self.kernel_size != 1:
            assert look_ahead >= 0, "look_ahead must be > 0."
            assert look_ahead < self.padding, "look_ahead must be < self.padding."

    def forward(self, x):
        """
        x: (B, T, D)
        y: (B, T, D)
        """
        x = x.transpose(1, 2)
        y = self.conv1d(x)
        # NOTE(k2kobayashi): kernel_size=1 does not discard padding
        if self.kernel_size != 1:
            if self.use_causal:
                y = y[..., self.look_ahead : -self.padding + self.look_ahead]
            else:
                y = y[..., self.padding // 2 : -self.padding // 2]
        return y.transpose(1, 2)


class ConvLayers(nn.Module):
    def __init__(
        self,
        in_channels,
        conv_channels=256,
        out_channels=222,
        kernel_size=3,
        dilation_size=1,
        group_size=8,
        look_ahead=0,
        use_causal=False,
        use_ddsconv=False,
        use_quefrency_norm=True,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.conv_channels = conv_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.dilation_size = dilation_size
        self.group_size = group_size
        self.use_causal = use_causal
        self.use_quefrency_norm = use_quefrency_norm
        self.look_ahead = look_ahead
        self.use_ddsconv = use_ddsconv

        if not use_ddsconv:
            self.net = self.original_conv()
        else:
            self.net = self.dds_conv()

        if self.use_quefrency_norm:
            idx = torch.arange(1, out_channels // 2 + 1).float()
            self.quef_norm = torch.cat([torch.flip(idx, dims=[-1]), idx], dim=-1)

    def forward(self, x):
        """
        x: (B, T, in_channels)
        y_comp: (B, T, out_channels)
        """
        y = self.net(x)
        if self.use_quefrency_norm:
            y = y / self.quef_norm.to(y.device)
        return y

    def original_conv(self):
        return nn.Sequential(
            Conv1d(
                self.in_channels,
                self.conv_channels,
                self.kernel_size,
                self.dilation_size,
                1,
                self.use_causal,
            ),
            nn.ReLU(),
            Conv1d(
                self.conv_channels,
                self.conv_channels,
                self.kernel_size,
                self.dilation_size,
                self.group_size,
                self.use_causal,
            ),
            nn.ReLU(),
            Conv1d(
                self.conv_channels,
                self.conv_channels,
                self.kernel_size,
                self.dilation_size,
                self.group_size,
                self.use_causal,
            ),
            nn.ReLU(),
            Conv1d(
                self.conv_channels,
                self.out_channels,
                self.kernel_size,
                self.dilation_size,
                1,
                self.use_causal,
            ),
        )

    def dds_conv(self):
        modules = []
        modules += [
            Conv1d(
                in_channels=self.in_channels,
                out_channels=self.conv_channels,
                kernel_size=1,
                dilation_size=1,
                group_size=1,
                use_causal=self.use_causal,
            )
        ]
        for i in range(3):
            if self.dilation_size == 1:
                dilation_size = self.kernel_size ** i
            else:
                dilation_size = self.dilation_size ** i
            modules += [
                DepthSeparableConvBlock(
                    channels=self.conv_channels,
                    kernel_size=self.kernel_size,
                    dilation_size=dilation_size,
                    use_causal=self.use_causal,
                )
            ]
        modules += [
            Conv1d(
                in_channels=self.conv_channels,
                out_channels=self.out_channels,
                kernel_size=1,
                dilation_size=1,
                group_size=1,
                use_causal=self.use_causal,
                look_ahead=self.look_ahead,
            )
        ]
        return nn.Sequential(*modules)


class DepthSeparableConvBlock(nn.Module):
    def __init__(self, channels, kernel_size, dilation_size, use_causal=False):
        super().__init__()
        self.sep_conv = Conv1d(
            channels,
            channels,
            kernel_size,
            dilation_size,
            group_size=channels,
            use_causal=use_causal,
        )
        self.conv1d = Conv1d(
            channels,
            channels,
            kernel_size=1,
            dilation_size=1,
            group_size=1,
            use_causal=use_causal,
        )
        self.ln1 = nn.LayerNorm(channels)
        self.ln2 = nn.LayerNorm(channels)

    def forward(self, x):
        y = self.sep_conv(x)
        y = self.ln1(y)
        y = F.gelu(y)
        y = self.conv1d(y)
        y = self.ln2(y)
        y = F.gelu(y)
        return x + y


class ConvLTVFilterGenerator(nn.Module):
    def __init__(
        self,
        in_channels,
        conv_channels=256,
        ltv_out_channels=222,
        kernel_size=3,
        dilation_size=1,
        group_size=8,
        fft_size=1024,
        hop_size=256,
        look_ahead=256,
        mel2spc_fn=None,
        use_causal=False,
        use_ddsconv=False,
        use_ltv_conv_postfilter=False,
        use_quefrency_norm=True,
    ):
        super().__init__()
        self.fft_size = fft_size
        self.hop_size = hop_size
        # TODO(k2kobayashi): support overlap size not only 50%
        self.window_size = hop_size * 2
        self.ltv_out_channels = ltv_out_channels
        self.mel2spc_fn = mel2spc_fn
        self.use_ltv_conv_postfilter = use_ltv_conv_postfilter

        win_norm = self.window_size // (hop_size * 2)  # only for hanning window
        self.win = torch.hann_window(self.window_size) / win_norm
        self.conv = ConvLayers(
            in_channels,
            conv_channels,
            ltv_out_channels,
            kernel_size=kernel_size,
            dilation_size=dilation_size,
            group_size=group_size,
            use_causal=use_causal,
            use_ddsconv=use_ddsconv,
            use_quefrency_norm=use_quefrency_norm,
        )

        if self.use_ltv_conv_postfilter:
            self.conv_pf = Conv1d(
                in_channels=1,
                out_channels=1,
                kernel_size=hop_size,
                use_causal=True,
                look_ahead=look_ahead,
            )

    def forward(self, x, z):
        """
        x: B, T, D
        z: B, 1, T * hop_size
        """
        ccep = self.conv(x)
        if self.mel2spc_fn is not None:
            log_mag = self.mel2spc_fn(x)
        else:
            log_mag = None

        y = self._ccep2impulse(ccep, ref=log_mag)
        z = self._conv_impulse(z, y)
        z = self._ola(z)
        if self.use_ltv_conv_postfilter:
            z = self.conv_pf(z.transpose(1, 2)).transpose(1, 2)

        z = torch.clamp(z, -1, 1)
        return z

    def _ccep2impulse(self, ccep, ref=None):
        padding = (self.fft_size - self.ltv_out_channels) // 2
        ccep = F.pad(ccep, (padding, padding))
        y = torch.fft.fft(ccep, n=self.fft_size, dim=-1)
        if ref is not None:
            # TODO(k2kobayashi): it requires to consider following line.
            # this mask eliminates very small amplitude values (-100).
            # ref = ref * (ref > -100)
            y.real[..., : self.fft_size // 2 + 1] += ref
            y.real[..., self.fft_size // 2 :] += torch.flip(ref[..., 1:], dims=[-1])

        # NOTE(k2kobayashi): we assume intermidiate log amplitude as 10log10|mag|
        mag, phase = torch.pow(10, y.real / 10), y.imag
        real, imag = mag * torch.cos(phase), mag * torch.sin(phase)
        y = torch.fft.ifft(torch.complex(real, imag), n=self.fft_size, dim=-1)
        return y.real

    def _conv_impulse(self, z, y):
        z = z.reshape(z.size(0), 1, -1)  # (B, 1, T x hop_size)
        z = F.pad(z, (self.window_size // 2 - 1, self.window_size // 2))
        z = z.unfold(-1, self.window_size, step=self.hop_size)  # (B, 1, T, window_size)

        z = F.pad(z, (self.fft_size // 2 - 1, self.fft_size // 2))
        z = z.unfold(-1, self.fft_size, step=1)  # (B, 1, T, window_size, fft_size)

        # z = matmul(z, y) -> (B, 1, T, window_size) where
        # z: (B, 1, T, window_size, fft_size)
        # y: (B, T, fft_size) -> (B, 1, T, fft_size, 1)
        z = torch.matmul(z, y.unsqueeze(1).unsqueeze(-1)).squeeze(-1)
        return z

    def _ola(self, z):
        z = z * self.win.to(z.device)
        l, r = torch.chunk(z, 2, dim=-1)  # (B, 1, T, window_size // 2)
        z = l + torch.roll(r, 1, dims=2)  # roll a frame of right side
        z = z.reshape(z.size(0), 1, -1)
        return z


class SinusoidsGenerator(nn.Module):
    def __init__(
        self,
        hop_size,
        fs=24000,
        excit_amp=0.1,
        noise_std=0.03,
        n_harmonics=200,
        use_uvmask=True,
    ):
        super().__init__()
        self.fs = fs
        self.excit_amp = excit_amp
        self.noise_std = noise_std
        self.upsample = nn.Upsample(scale_factor=hop_size, mode="nearest")
        self.use_uvmask = True

        self.n_harmonics = n_harmonics
        self.harmonics = torch.arange(1, self.n_harmonics + 1).unsqueeze(-1)

    def forward(self, cf0, uv):
        device = cf0.device
        f0, uv = self.upsample(cf0.transpose(1, 2)), self.upsample(uv.transpose(1, 2))
        excit = self.generate_sinusoids(f0, uv).reshape(cf0.size(0), 1, -1)
        noise = torch.normal(0, self.noise_std, size=excit.size()).to(device)
        return excit, noise

    def generate_sinusoids(self, f0, uv):
        harmonics = self.harmonics.to(f0.device)
        mask = self.anti_aliacing_mask(f0 * harmonics)
        rads = f0.cumsum(dim=-1) * 2.0 * math.pi / self.fs * harmonics
        excit = torch.sum(torch.cos(rads) * mask, dim=1, keepdim=True)
        if self.use_uvmask:
            excit = uv * excit
        return self.excit_amp * excit

    def anti_aliacing_mask(self, f0_with_harmonics, use_soft_mask=False):
        if use_soft_mask:
            return torch.sigmoid(-(f0_with_harmonics - self.fs / 2.0))
        else:
            return (f0_with_harmonics < self.fs / 2.0).float()
